/*
 * Copyright 2012 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.core;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

public class NodeDesc {
    private UUID uuid;
    private String id;
    private Map<String, Attribute> attributes;
    private List<UUID> childrenUUId;
    private Map<UUID, NodeDesc> children;
    private UUID parentUUId;

    public NodeDesc(UUID uuid, UUID parentUUId, String id,
            Map<String, Attribute> attributes, List<UUID> childrenUUId) {
        this.uuid = uuid;
        this.id = id;
        this.attributes = new HashMap<String, Attribute>();
        for (String key : attributes.keySet()) {
            Attribute attr = attributes.get(key);
            Object value = attr.getValue();
            if (value instanceof List) {
                value = new ArrayList<Object>((List<?>)value);
            }
            this.attributes.put(key, new Attribute(attr.getType(), attr.getKey(), value));
        }
        this.childrenUUId = new ArrayList<UUID>(childrenUUId);
        this.children = new HashMap<UUID, NodeDesc>();
        this.parentUUId = parentUUId;
    }

    public void setId(String id) {
        this.id = id;
    }

    @Override
    public boolean equals(Object obj) {
        if (!(obj instanceof NodeDesc)) {
            return false;
        }
        NodeDesc nodeDesc = (NodeDesc) obj;
        if (!nodeDesc.getId().equals(getId())) {
            return false;
        }
        if (!nodeDesc.getUUId().equals(getUUId())) {
            return false;
        }
        if (!attrubuteEquals(nodeDesc)) {
            return false;
        }
        if (!childrenEquals(nodeDesc)) {
            return false;
        }
        return true;
    }

    public boolean equalsButUUID(Object obj) {
        if (!(obj instanceof NodeDesc)) {
            return false;
        }
        NodeDesc nodeDesc = (NodeDesc) obj;
        if (!nodeDesc.getId().equals(getId())) {
            return false;
        }
        if (!attrubuteEquals(nodeDesc)) {
            return false;
        }
        if (!childrenEquals(nodeDesc)) {
            return false;
        }
        return true;
    }

    private boolean childrenEquals(NodeDesc nodeDesc) {
        return getChildrenUUId().equals(nodeDesc.getChildrenUUId());
    }

    private boolean attrubuteEquals(NodeDesc nodeDesc) {
        return nodeDesc.getAttributes().equals(getAttributes());
    }

    public String getId() {
        return id;
    }

    public UUID getUUId() {
        return uuid;
    }

    public List<UUID> getChildrenUUId() {
        return Collections.unmodifiableList(childrenUUId);
    }

    public Map<String, Attribute> getAttributes() {
        return attributes;
    }

    public void setAttribute(Attribute attr) {
        attributes.put(attr.getKey(), attr);
    }

    public UUID getParentUUId() {
        return parentUUId;
    }

    private String getType() {
        return getAttributes().get(TreeNode.ATTR_TYPE).getValue().toString();
    }

    public void setChildrenDesc(UUID chUUId, NodeDesc chDesc) {
        assert (childrenUUId.contains(chUUId));
        children.put(chUUId, chDesc);
    }

    public NodeDesc getChildDesc(UUID chUUId) {
        return children.get(chUUId);
    }

    public TreeNode createNode(TreeDB db, UUID pUUId, Boolean deep) {
        // new node - new uuid
        TreeNode result = db.createChildNode(db.getUuidGenerator().nextUUId(),
                pUUId, id, getType());
        for (Attribute a : getAttributes().values()) {
            result.putAttribute(a);
        }
        result.saveAttributes();

        if (deep) {
            for (UUID chUUId : getChildrenUUId()) {
                children.get(chUUId).createNode(db, result.getUUId(), deep);
            }
        }
        return result;
    }

    public TreeNode createNode(TreeDB db) {
        return createNode(db, parentUUId, true);
    }

    public void setParentUUId(UUID pUUId) {
        this.parentUUId = pUUId;
    }

    static public NodeDesc fromNode(TreeNode node) {
        return fromNode(node, true);
    }

    static public NodeDesc fromNode(TreeNode node, Boolean deep) {
        NodeDesc result = new NodeDesc(node.getUUId(), node.getParentUUId(),
                node.getId(), node.getAttributes(), node.getChildrenUUIds());
        if (deep) {
            for (int i = 0; i < node.sizeChildren(); i++) {
                result.setChildrenDesc(node.getChild(i).getUUId(), node
                        .getChild(i).toNodeDesc());
            }
        }
        return result;
    }
}