/*
 * Copyright 2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.core;

import java.io.File;
import java.io.InputStream;
import java.net.MalformedURLException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.Comparator;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.UUID;

public class TreeNode {
    public static final String ATTR_TYPE = "_type";
    public static final String ATTR_PREDICATE = "_predicate";
    public static final String ATTR_DESCRIPTION = "_description";
    public static final String ATTR_BASE_NODES = "_baseNodes";

    private UUID uuid;
    private final TreeDB db;
    protected String id;
    UUID parent;
    List<UUID> children = new ArrayList<UUID>();
    Map<String, Attribute> attributes = new LinkedHashMap<String, Attribute>();

    public TreeNode(TreeDB db, UUID uuid) {
        if (uuid == null)
            uuid = UUID.randomUUID();
        this.uuid = uuid;
        this.db = db;
    }

    public void setPredicate(String pred) {
        putAttribute(new Attribute(AttributeType.STRING, ATTR_PREDICATE, pred));
    }

    public String getPredicate() {
        if (getAttributeValue(ATTR_PREDICATE) == null) {
            return "";
        }
        return (String) getAttributeValue(ATTR_PREDICATE);
    }

    public void setType(String type) {
        putAttribute(new Attribute(AttributeType.STRING, ATTR_TYPE, type));
    }

    public final UUID getUUId() {
        return uuid;
    }

    public void setId(String id) {
        db.setId(this, id);
    }

    public String getId() {
        return id;
    }

    /**
     * @return String name of node type
     */
    public String getType() {
        return (String) getAttributeValue(ATTR_TYPE);
    }

    public void setParent(TreeNode parent) {
        db.moveNode(this, parent);
    }

    public UUID getParentUUId() {
        return parent;
    }

    public void removeChild(TreeNode node) {
        db.removeNode(node);
    }

    public void setDescription(String description) {
        putAttribute(new Attribute(AttributeType.STRING,
                ATTR_DESCRIPTION, description));
    }

    public String getDescription() {
        return getDescription(false);
    }

    public String getDescription(boolean process) {
        if (attributes.get(ATTR_DESCRIPTION) != null) {
            String result = (String) attributes.get(ATTR_DESCRIPTION)
                    .getValue();
            if (process) {
                result = processDescription(result);
            }
            return result;
        }
        return null;
    }

    private String processDescription(String desc) {
        String resUrl = getResourcesUrl().toString();
        String result = desc.replace("${node.resURL}", resUrl);
        return result;
    }

    public TreeDB getTreeDB() {
        return db;
    }

    void setUUId(UUID uuid) {
        this.uuid = uuid;
    }

    public TreeNode getParent() {
        if (parent == null) {
            // no parent
            assert this.getTreeDB().getRootNode().equals(this);
            return null;
        }
        return db.getNode(parent);
    }

    public int sizeChildren() {
        return children.size();
    }

    public TreeNode getChild(int i) {
        return db.getNode(children.get(i));
    }

    public String getQualifiedId() {
        if (getParent() == null) {
            return "/";
        }
        String parentQId = getTreeDB().getNode(parent).getQualifiedId();
        if (parentQId.equals("/")) {
            return parentQId + this.getId();
        }
        return parentQId + "/" + this.getId();
    }

    public TreeNode findChild(String id) {
        for (int i = 0; i < sizeChildren(); i++) {
            if (getChild(i).getId().equals(id)) {
                return getChild(i);
            }
        }
        return null;
    }

    public List<UUID> getChildrenUUIds() {
        return Collections.unmodifiableList(children);
    }

    @Override
    public String toString() {
        return getId() + "(" + getType() + ")";
    }

    public TreeNode createChild(String id) {
        if (findChild(id) != null)
            throw new IllegalArgumentException("child '" + id
                    + "' is already exists");
        if (getType().equals("DBRoot")) {
            return db.createChildNode(uuid, id, "Node");
        }
        return db.createChildNode(uuid, id, getType());
    }

    public TreeNode createChild(String id, String type) {
        return db.createChildNode(uuid, id, type);
    }

    public void delete() {
        getParent().removeChild(this);
    }

    public Attribute putAttribute(Attribute a) {
        Attribute result = attributes.get(a.getKey());
        attributes.put(a.getKey(), a);
        return result;
    }

    public Attribute removeAttribute(String key) {
        return attributes.remove(key);
    }

    public Attribute getAttribute(String key) {
        return attributes.get(key);
    }

    public Object getAttributeValue(String key) {
        return getAttributeValue(key, null);
    }

    public Object getAttributeValue(String key, Object def) {
        Attribute value = attributes.get(key);
        return value != null ? value.getValue() : def;
    }

    public int getIntAttributeValue(String key, int def) {
        Attribute value = attributes.get(key);
        return value != null ? (Integer) value.getValue() : def;
    }

    public int getIntAttributeValue(String key) {
        return getIntAttributeValue(key, 0);
    }

    public float getFloatAttributeValue(String key, float def) {
        Attribute value = attributes.get(key);
        return value != null ? (Float) value.getValue() : def;
    }

    public float getFloatAttributeValue(String key) {
        return getFloatAttributeValue(key, 0);
    }

    public boolean getBooleanAttributeValue(String key, boolean def) {
        Attribute value = attributes.get(key);
        return value != null ? (Boolean) value.getValue() : def;
    }

    public boolean getBooleanAttributeValue(String key) {
        return getBooleanAttributeValue(key, false);
    }

    public String getStringAttributeValue(String key, String def) {
        Attribute value = attributes.get(key);
        return value != null ? (String) value.getValue() : def;
    }

    public String getStringAttributeValue(String key) {
        return getStringAttributeValue(key, null);
    }

    @SuppressWarnings("unchecked")
    public List<String> getListAttributeValue(String key) {
        return (List<String>) getAttributeValue(key, null);
    }

    public Set<String> getAttributeKeys() {
        return Collections.unmodifiableSet(attributes.keySet());
    }

    public Map<String, Attribute> getAttributes() {
        return Collections.unmodifiableMap(attributes);
    }

    public void saveAttributes() {
        db.saveAttributes(uuid, this.attributes.values());
    }

    @SuppressWarnings("unchecked")
    public void addBaseNode(String nodeQId) {
        List<String> baseNodes = (List<String>) getAttributeValue(ATTR_BASE_NODES);
        if (baseNodes == null) {
            baseNodes = new ArrayList<String>();
        }
        baseNodes.add(nodeQId);
        putAttribute(new Attribute(AttributeType.LIST, ATTR_BASE_NODES,
                baseNodes));
    }

    @SuppressWarnings("unchecked")
    public boolean removeBaseNode(String nodeQId) {
        List<String> baseNodes = (List<String>) getAttributeValue(ATTR_BASE_NODES);
        if (baseNodes == null || !baseNodes.remove(nodeQId)) {
            return false;
        }
        putAttribute(new Attribute(AttributeType.LIST, ATTR_BASE_NODES,
                baseNodes));
        return true;
    }

    @SuppressWarnings("unchecked")
    public List<String> getBaseNodes() {
        if (getAttributeValue(ATTR_BASE_NODES) == null) {
            return Collections.EMPTY_LIST;
        }
        return Collections
                .unmodifiableList((List<String>) getAttributeValue(ATTR_BASE_NODES));
    }

    @Override
    public boolean equals(Object obj) {
        if (obj == null)
            return false;
        if (super.equals(obj))
            return true;
        if (!(obj instanceof TreeNode))
            return false;
        TreeNode node = (TreeNode) obj;
        if (!node.getUUId().equals(uuid))
            return false;
        if (!getQualifiedId().equals(node.getQualifiedId()))
            return false;
        if (!getAttributeKeys().equals(node.getAttributeKeys()))
            return false;
        for (String key : getAttributeKeys()) {
            if (!getAttributeValue(key).equals(node.getAttributeValue(key)))
                return false;
        }
        return true;
    }

    public TreeNode[] getSortedChildren() {
        return getSortedChildren(null, null);
    }

    public TreeNode[] getSortedChildren(String type) {
        return getSortedChildren(null, type);
    }

    public TreeNode[] getSortedChildren(Comparator<TreeNode> c, String type) {
        ArrayList<TreeNode> result = new ArrayList<TreeNode>();
        for (int i = 0; i < sizeChildren(); i++) {
            TreeNode ch = getChild(i);
            if (type == null || ch.getType().equals(type)) {
                result.add(getChild(i));
            }
        }
        TreeNode[] r = (TreeNode[]) result.toArray(new TreeNode[0]);
        if (c != null) {
            Arrays.sort(r, c);
        } else {
            Arrays.sort(r, new Comparator<TreeNode>() {
                @Override
                public int compare(TreeNode o1, TreeNode o2) {
                    int result = 0;
                    try {
                        if (o1 == null || o2 == null) {
                            // FIXME: why o0 or o2 can be null?
                            return 0;
                        }
                        Integer i1 = Integer.valueOf(o1.getAttributeValue(
                                TreeDB.ORDER_ATTRIBUTE, 0).toString());
                        Integer i2 = Integer.valueOf(o2.getAttributeValue(
                                TreeDB.ORDER_ATTRIBUTE, 0).toString());

                        if ((result = i1.compareTo(i2)) == 0) {
                            return o1.getId().compareTo(o2.getId());
                        }
                    } catch (NumberFormatException e) {
                        // ordering attribute can be with errors
                        // e.printStackTrace();
                    }
                    return result;
                }
            });
        }
        return r;
    }

    public URL getResourcesUrl() {
        try {
            return new File(db.getStorage().getNodeResourcePath(uuid)).toURI()
                    .toURL();
        } catch (MalformedURLException e) {
            e.printStackTrace();
        }
        return null;
    }

    public NodeDesc toNodeDesc() {
        return toNodeDesc(true);
    }

    public NodeDesc toNodeDesc(Boolean deep) {
        return NodeDesc.fromNode(this, deep);
    }

    public InputStream getResourceContent(String name) {
        return db.getStorage().getResourceContents(uuid, name);
    }

    public void setResourceContent(String name, InputStream in) {
        db.setResourceContent(this, name, in, null);
    }

    public void setResourceContent(String name, InputStream in, Object data) {
        db.setResourceContent(this, name, in, data);
    }

    public void deleteResource(String name) {
        db.deleteResource(this, name);
    }

    public boolean isResourceExists(String name) {
        return db.isResourceExists(this, name);
    }
}
