/*
 * Copyright 2009-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.unitesk.requality.eclipse.core;

import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.CopyOnWriteArrayList;

import org.eclipse.core.commands.operations.IOperationHistory;
import org.eclipse.core.commands.operations.IUndoContext;
import org.eclipse.core.resources.IFile;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IncrementalProjectBuilder;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.resources.WorkspaceJob;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.ui.PlatformUI;

import com.unitesk.requality.core.NodeDesc;
import com.unitesk.requality.core.TreeDB;
import com.unitesk.requality.core.TreeNode;
import com.unitesk.requality.core.transaction.ITransactionListener;
import com.unitesk.requality.core.transaction.TransactionStorage;
import com.unitesk.requality.eclipse.Activator;
import com.unitesk.requality.eclipse.NodeTypeDescriptor;
import com.unitesk.requality.eclipse.RequalityPlugin;
import com.unitesk.requality.eclipse.natures.ReqCatalogue;
import com.unitesk.requality.eclipse.operations.TreeDBOperation;

public class TreesTracker {
    public enum TreeEvent {
        TREE_OPEN, TREE_CLOSE, TREE_DELETE
    }

    private Map<IProject, TreeDB> treeByProject = new HashMap<IProject, TreeDB>();
    private Map<TreeDB, IProject> projectByTree = new HashMap<TreeDB, IProject>();
    private Map<IProject, ResourceStorage> projectToStorage = new HashMap<IProject, ResourceStorage>();

    private static TreesTracker tracker;
    private List<ITreeStateListener> listeners = new CopyOnWriteArrayList<ITreeStateListener>();

    public boolean addListener(ITreeStateListener l) {
        return listeners.add(l);
    }

    public boolean removeListener(ITreeStateListener l) {
        return listeners.remove(l);
    }

    private TreesTracker() {
    }

    public static TreesTracker getTracker() {
        if (tracker == null)
            tracker = new TreesTracker();
        return tracker;
    }

    public synchronized void notifyListeners(TreeDB db, TreeEvent type) {
        for (ITreeStateListener listener : listeners) {
            switch (type) {
            case TREE_CLOSE:
                listener.close(db);
                break;
            case TREE_OPEN:
                listener.open(db);
                break;
            case TREE_DELETE:
                listener.delete(projectByTree.get(db));
                break;
            }
        }
    }

    public IProject getProject(TreeDB db) {
        return projectByTree.get(db);
    }

    public TreeDB getTree(IProject proj) {
        return treeByProject.get(proj);
    }

    public boolean trackProject(IProject project) {
        try {
            if (project.isOpen() && project.isNatureEnabled(ReqCatalogue.ID)) {
                TreeDB db = getTree(project);
                if (db == null) {
                    db = loadTreeDB(project);
                    Map<String, NodeTypeDescriptor> nts = RequalityPlugin
                            .getAdditionalTypes();
                    for (String name : RequalityPlugin.getAdditionalTypes()
                            .keySet()) {
                        db.addNodeType(name, nts.get(name).getClazz());
                    }
                    bindBaseToProject(project, db);
                    db.addListener(new DefaultTreeChangeListener() {
                        @Override
                        public void deleted(NodeDesc node) {
                            // TODO: is it right place for this?
                            String type = node.getAttributes().get("_type")
                                    .getValue().toString();
                            if (Activator.getDefault().getActiveNode(type) == null) {
                                return;
                            }
                            if (! Activator.getDefault().getActiveNode(type)
                                    .equals(node)) {
                                Activator.getDefault()
                                        .setActiveNode(null, type);
                            }
                        }
                    });
                    notifyListeners(db, TreeEvent.TREE_OPEN);
                }
            } else
                return false;
        } catch (Exception e) {
            e.printStackTrace();
            return false;
        }
        return true;
    }

    private TreeDB loadTreeDB(final IProject project) {
        if(!projectToStorage.containsKey(project)) {
            projectToStorage.put(project, new ResourceStorage(project));
        }
        ResourceStorage storage = projectToStorage.get(project);
        TreeDB db = TreeDB.load(storage);
        db.addTransactionListener(new ITransactionListener() {
            public IUndoContext getContext() {
                return PlatformUI.getWorkbench()
                        .getOperationSupport().getUndoContext();
            }

            @Override
            public void dispose() {
                IOperationHistory history = PlatformUI.getWorkbench()
                        .getOperationSupport().getOperationHistory();
                history.dispose(getContext(), true, true, true);
            }

            @Override
            public void execute(TransactionStorage tStorage) {
                TreeDBOperation operation = new TreeDBOperation(tStorage);

                operation.addContext(getContext());
                IOperationHistory history = PlatformUI.getWorkbench()
                        .getOperationSupport().getOperationHistory();
                history.add(operation);
            }

            @Override
            public void redo(TransactionStorage tStorage) {
            }

            @Override
            public void undo(TransactionStorage tStorage) {
                // TODO Auto-generated method stub

            }
        });

        ResourcesPlugin.getWorkspace().addResourceChangeListener(
                storage.getResourceChangeListener());
        WorkspaceJob runBuilder = new WorkspaceJob("Check Requality Project: "
                + project.getName()) {
            @Override
            public IStatus runInWorkspace(IProgressMonitor monitor)
                    throws CoreException {
                try {
                    project.build(IncrementalProjectBuilder.FULL_BUILD, null);
                } catch (CoreException e) {
                    e.printStackTrace();
                }
                return Status.OK_STATUS;
            }
        };
        runBuilder.setRule(ResourcesPlugin.getWorkspace().getRuleFactory().buildRule());
        runBuilder.schedule();
        return db;
    }

    public boolean stopTracking(IProject project, boolean deleted) {
        if (!isTracking(project)) {
            return false;
        }
        TreeDB db = getTree(project);

        ResourceStorage storage = (ResourceStorage) db.getStorage();
        ResourcesPlugin.getWorkspace().removeResourceChangeListener(
                storage.getResourceChangeListener());

        projectToStorage.remove(project);
        if (!deleted) {
            notifyListeners(db, TreeEvent.TREE_CLOSE);
            unbindBaseFromProject(project, db);
            treeByProject.put(project, null);
        } else {
            notifyListeners(db, TreeEvent.TREE_DELETE);
            unbindBaseFromProject(project, db);
        }
        return true;
    }

    private void unbindBaseFromProject(IProject project, TreeDB db) {
        projectByTree.remove(db);
        treeByProject.remove(project);
    }

    private void bindBaseToProject(IProject project, TreeDB db) {
        projectByTree.put(db, project);
        treeByProject.put(project, db);
    }

    public boolean isTracking(IProject project) {
        boolean result = treeByProject.containsKey(project);
        return result;
    }

    public TreeDB[] getTrees() {
        return projectByTree.keySet().toArray(new TreeDB[0]);
    }

    public IProject[] getProjects() {
        return projectByTree.values().toArray(new IProject[0]);
    }

    private boolean isTrackable(IProject project) {
        try {
            return project.isNatureEnabled(ReqCatalogue.ID);
        } catch (Exception e) {
            e.printStackTrace();
            return false;
        }
    }

    public void setProjects(IProject[] projects) {
        HashSet<IProject> current = new HashSet<IProject>(
                treeByProject.keySet());

        for (IProject project : projects) {

            try {
                if (isTrackable(project) && !isTracking(project)) {
                    trackProject(project);
                } else if (!isTrackable(project) && isTracking(project)) {
                    stopTracking(project, project.exists());
                }
                current.remove(project);
            } catch (Exception e) {
                e.printStackTrace();
            }
        }

        for (IProject iProject : current) {
            stopTracking(iProject, iProject.exists());
        }
    }

    public static TreeNode getNodeByResource(IFile file) {
        TreeDB db = getTracker().getTree(file.getProject());
        ResourceStorage rs = getResourceStorage(db);
        UUID uuid = rs.getUUId(file);
        if (uuid == null) {
            return null;
        }
        return db.getNode(uuid);
    }

    public static IResource getResourceByNode(TreeNode node) {
        return getResourceStorage(node.getTreeDB()).getResource(node.getUUId());
    }

    public static ResourceStorage getResourceStorage(TreeDB db) {
        ResourceStorage rs;
        if (db.getStorage() instanceof TransactionStorage) {
            rs = (ResourceStorage) ((TransactionStorage) db.getStorage())
                    .getStorage();
        } else {
            rs = (ResourceStorage) db.getStorage();
        }
        return rs;
    }

    public void assignTreeDB(IProject project, ResourceStorage resourceStorage) {
        projectToStorage.put(project, resourceStorage);
    }
}
